from PyQt5 import QtGui, QtWidgets
from PyQt5.QtWidgets import *
from PyQt5.QtGui import *
from PyQt5.QtCore import pyqtSignal, QThread, Qt, QEvent

import difflib
import sys
import json

import settings

class GUI(QtWidgets.QMainWindow):
    def __init__(self, parent = None):
        QtWidgets.QMainWindow.__init__(self,parent)
        self.initUI()
        self.drawCorrections()

    def initToolbar(self):
        self.openAction = QtWidgets.QAction(QtGui.QIcon.fromTheme("document-open"),"Open",self)
        self.openAction.setStatusTip("Open document")
        self.openAction.setShortcut("Ctrl+O")
        self.openAction.triggered.connect(self.openDocument)
        self.toolbar = self.addToolBar("Options")
        self.toolbar.addAction(self.openAction)

        self.saveAction = QtWidgets.QAction(QtGui.QIcon.fromTheme("document-save"),"Save",self)
        self.saveAction.setStatusTip("Save document")
        self.saveAction.setShortcut("Ctrl+S")
        self.saveAction.triggered.connect(self.save)
        self.toolbar.addAction(self.saveAction)

        self.spellcheckAction = QtWidgets.QAction(QtGui.QIcon.fromTheme("tools-check-spelling"),"Spellcheck",self)
        self.spellcheckAction.setStatusTip("Spell-check document")
        self.spellcheckAction.setShortcut("Ctrl+R")
        self.spellcheckAction.triggered.connect(self.recheck)
        self.toolbar.addAction(self.spellcheckAction)

        self.prevAction = QtWidgets.QAction(QtGui.QIcon.fromTheme("go-previous"),"Previous error",self)
        self.prevAction.setStatusTip("Go to previous error")
        self.prevAction.setShortcut("Ctrl+Up")
        self.prevAction.triggered.connect(self.gotoPrevError)
        self.toolbar.addAction(self.prevAction)

        self.nextAction = QtWidgets.QAction(QtGui.QIcon.fromTheme("go-next"),"Next error",self)
        self.nextAction.setStatusTip("Go to next error")
        self.nextAction.setShortcut("Ctrl+Down")
        self.nextAction.triggered.connect(self.gotoNextError)
        self.toolbar.addAction(self.nextAction)

        self.settingsAction = QtWidgets.QAction(QtGui.QIcon.fromTheme("preferences-other"),"Settings",self)
        self.settingsAction.setStatusTip("Settings")
        self.settingsAction.setShortcut("Ctrl+P")
        self.settingsAction.triggered.connect(lambda x: SettingsDialog())
        self.toolbar.addAction(self.settingsAction)

        self.zoomInAction = QtWidgets.QAction(QtGui.QIcon.fromTheme("zoom-in"),"+",self)
        self.zoomInAction.setStatusTip("Zoom in")
        self.zoomInAction.setShortcut("Ctrl++")
        self.zoomInAction.triggered.connect(self.text.zoomIn)
        self.toolbar.addAction(self.zoomInAction)

        self.zoomOutAction = QtWidgets.QAction(QtGui.QIcon.fromTheme("zoom-out"),"-",self)
        self.zoomOutAction.setStatusTip("Zoom out")
        self.zoomOutAction.setShortcut("Ctrl+-")
        self.zoomOutAction.triggered.connect(self.text.zoomOut)
        self.toolbar.addAction(self.zoomOutAction)


    def setCursor(self, pos):
        cursor = self.text.textCursor()
        cursor.setPosition( pos, QTextCursor.MoveAnchor)
        self.text.setTextCursor(cursor)
        self.text.setFocus()


    def initUI(self):
        self.text = SpellTextEdit(self)
        self.lasttext = ""
        self.initToolbar()
        #self.text.setTabStopWidth(33)
        #self.setCentralWidget(self.text)

        self.explain = QTextEdit(self)
        self.explain.setHtml("No errors found")


        split = QSplitter(self)

        grid = QtWidgets.QGridLayout()
        grid.setSpacing(10)

        grid.addWidget(self.explain, 1, 0)

        #grid.setColumnStretch(0, 3)
        #grid.setColumnStretch(1, 1)

        sidebar = QWidget(self)
        sidebar.setLayout(grid)

        split.addWidget(self.text)
        split.addWidget(sidebar)
        split.setStretchFactor(0, 3)
        split.setStretchFactor(1, 1)


        self.progressbar = QWidget(self)
        grid2 = QtWidgets.QGridLayout(self)
        grid2.setSpacing(10)

        self.status = QLabel("Status")
        grid2.addWidget(self.status, 1, 0)

        self.progress = QtWidgets.QProgressBar(self)
        grid2.addWidget(self.progress, 2, 0)

        self.progressbar.setLayout(grid2)
        split2 = QSplitter(self)
        split2.addWidget(split)
        split2.addWidget(self.progressbar)
        split2.setOrientation(Qt.Vertical)

        self.progressbar.setVisible(False)

        self.setCentralWidget(split2)

        self.setGeometry(100,100,1030,800)
        self.setWindowTitle("LaTeX Grammarly")
        self.setWindowIcon(QtGui.QIcon("icons/icon.png"))

        self.text.textChanged.connect(self.textupdate)

        for i in range(9):
            action = QtWidgets.QAction(('Correct'), self)
            action.setShortcut("Ctrl+%d" % (i + 1))
            self.text.addAction(action)
            action.triggered.connect(lambda a, x=i: self.applyCorrection(x))





    def drawCorrections(self):
        self.lasttext = settings.text
        self.text.setPlainText(settings.text)

        cursor = self.text.textCursor()
        newcharfmt = QTextCharFormat()
        newcharfmt.setFontUnderline( True )
        newcharfmt.setUnderlineColor( Qt.red)
        newcharfmt.setUnderlineStyle( QTextCharFormat.WaveUnderline )
        newcharfmt.setBackground( Qt.yellow )
        newcharfmt.setFontWeight( QFont.Bold )

        mark = []
        for i in settings.correction.replace_list:
            mark.append((i, settings.correction.end(i)))

            cursor.setPosition( i + 1, QTextCursor.MoveAnchor)
            l = 1
            if settings.correction.end(i) - i < 2: l = 0
            cursor.setPosition( settings.correction.end(i) - l + 1, QTextCursor.KeepAnchor)
            self.text.setTextCursor(cursor)

            self.text.setCurrentCharFormat( newcharfmt )

            #self.text.setFocus()

        if len(mark) > 0:
            settings.correction.goto_closest_next(0)
            self.setCursor(settings.correction.to_pos() + 1)
            self.show_corrections()




    def applyCorrection(self, nr):
        has, l = self.text.highlightCorrection()
        text = self.text.textCursor().selectedText()
        if has:
            try:
                r = settings.correction.current_obj()[0]["replacements"][nr]
                if len(text) > 0 and l == 0: r += text[-1]
                self.text.correctWord(r)
                settings.correction.remove(settings.correction.current_id())
            except:
                pass


    def textupdate(self):
        text = self.text.toPlainText()

        check_len = 128

        cursor = self.text.textCursor()
        check_from = max(0, cursor.position() - int(check_len / 2))
        check_to = min(cursor.position() + int(check_len / 2), len(text) - 1)

        check_text = text[check_from:check_to]
        check_lasttext = self.lasttext[check_from:check_to]

        #print("Compare: ")
        #print("  > %s" % check_text)
        #print("  < %s" % check_lasttext)

        s = difflib.SequenceMatcher(None, check_lasttext, check_text)
        for tag, i1, i2, j1, j2 in s.get_opcodes():
            if tag != "equal":
                if i2 == check_len or j2 == check_len:
                    continue
                #print("%7s a[%d:%d] (%s) b[%d:%d] (%s)" % (tag, i1, i2, check_lasttext[i1:i2], j1, j2, check_text[j1:j2]))
                oldlen = i2 - i1
                newlen = j2 - j1
                delta = newlen - oldlen
                #print("Delta: %d" % delta)
                settings.correction.correct(i1 + check_from, delta)

        self.lasttext = text


    def recheck(self):
        settings.real_check = True
        self.thread = ParseThread(self)
        self.thread.starting.connect(lambda x: self.progressbar.setVisible(True))
        self.thread.update_status.connect(lambda x: self.status.setText(x))
        self.thread.update_progress.connect(lambda x: self.progress.setValue(x))
        self.thread.done.connect(lambda x: [self.drawCorrections(), self.progressbar.setVisible(False)])
        self.thread.start()


    def save(self):
        with open(sys.argv[1], "w") as w:
            w.write(self.text.toPlainText())
        settings.correction.to_file("corrections.json")
        
    def openDocument(self):
        fname = QFileDialog.getOpenFileName(self, 'Open file', '', "LaTeX files (*.tex)")
        if fname != None:
            fname = fname[0]
            if fname != "":
                if len(sys.argv) < 2:
                    sys.argv.append("")
                sys.argv[1] = fname;
                self.recheck()


    def gotoNextError(self):
        settings.correction.goto_closest_next(self.text.textCursor().position())
        self.setCursor(settings.correction.to_pos() + 1)
        self.show_corrections()


    def gotoPrevError(self):
        settings.correction.goto_closest_prev(self.text.textCursor().position())
        self.setCursor(settings.correction.to_pos() + 1)
        self.show_corrections()


    def show_corrections(self):
        c = settings.correction.current_obj()
        html = "<style>body { font-family: Verdana, Arial; }</style>"
        if c:
            if "title" in c[0]: html += c[0]["title"]
            if "explanation" in c[0]: html += "<span style='font-size:80%;'>" + c[0]["explanation"] + "</span>"
            if "replacements" in c[0]:
                html += "<ol>"
                for idx,r in enumerate(c[0]["replacements"]):
                    replace = r
                    if replace == "": replace = "<del>" + c[0]["text"] + "</del>"
                    html += "<li>%s</li>" % (replace)
                html += "</ol>"
        self.explain.setHtml(html)


    def clear_corrections(self):
        self.explain.setHtml("")



class SettingsDialog(QDialog):

    def __init__(self, *args):
        QDialog.__init__(self, *args)
        self.initUI()


    def initUI(self):
        main = QWidget(self)
        main.resize(400,480)
        grid = QtWidgets.QGridLayout(self)
        grid.setSpacing(10)

        grid.addWidget(QLabel("User Name"), 1, 0)
        self.username = QLineEdit()
        self.username.setText(settings.form_data["User"])
        grid.addWidget(self.username, 1, 1)

        grid.addWidget(QLabel("Password"), 2, 0)
        self.password = QLineEdit()
        self.password.setEchoMode(QLineEdit.Password)
        self.password.setText(settings.form_data["Pwd"])
        grid.addWidget(self.password, 2, 1)

        g1 = QGroupBox("Dialect")
        self.dialect_american = QRadioButton("American")
        self.dialect_british = QRadioButton("British")
        vbox = QHBoxLayout()
        vbox.addWidget(self.dialect_american)
        vbox.addWidget(self.dialect_british)
        vbox.addStretch(1)
        g1.setLayout(vbox)

        grid.addWidget(g1, 3, 0, 1, 2)

        self.dialect_american.setChecked(settings.form_data["Dialect"] == 1)
        self.dialect_british.setChecked(settings.form_data["Dialect"] == 2)

        self.genre = QComboBox(self)
        for g in settings.genres:
            self.genre.addItem(g["name"])
        self.genre.setCurrentIndex(settings.form_data["Genre"])

        grid.addWidget(QLabel("Genre"), 4, 0)
        grid.addWidget(self.genre, 4, 1)

        self.parsing = QComboBox(self)
        for p in settings.parsing_start:
            self.parsing.addItem(p[0])
        self.parsing.setCurrentIndex(settings.form_data["Parsing"])

        grid.addWidget(QLabel("Start parsing"), 5, 0)
        grid.addWidget(self.parsing, 5, 1)

        g2 = QGroupBox("Warn on")
        vbox = QVBoxLayout()
        self.warn_cat = []
        for i, e in enumerate(settings.err_cat):
            cb = QCheckBox(e["name"])
            cb.setChecked(settings.form_data["cat%d" % i])
            self.warn_cat.append(cb)
            vbox.addWidget(cb)
        vbox.addStretch(1)
        g2.setLayout(vbox)
        grid.addWidget(g2, 6, 0, 1, 2)

        self.save = QPushButton("Save")
        self.cancel = QPushButton("Cancel")

        grid.addWidget(self.cancel, 7, 0)
        grid.addWidget(self.save, 7, 1)

        main.setLayout(grid)

        self.save.clicked.connect(self.save_settings)
        self.cancel.clicked.connect(self.close)

        self.setWindowTitle("Settings")
        self.setWindowModality(Qt.ApplicationModal)
        self.exec_()


    def save_settings(self):
        settings.form_data["User"] = self.username.text()
        settings.form_data["Pwd"] = self.password.text()
        settings.form_data["Dialect"] = 1 if self.dialect_american.isChecked() else 2
        settings.form_data["Genre"] = self.genre.currentIndex()
        settings.form_data["Parsing"] = self.parsing.currentIndex()
        for i, e in enumerate(settings.err_cat):
            settings.form_data["cat%d" % i] = self.warn_cat[i].isChecked()
        with open("settings.json", "w") as s:
            json.dump(settings.form_data, s, indent = 2, sort_keys = True)
        self.close()
 

class SpellTextEdit(QPlainTextEdit):

    def __init__(self, *args):
        QPlainTextEdit.__init__(self, *args)
        self.gui = args[0]

        font = QFont("monospace");
        font.setStyleHint(QFont.Monospace);
        self.setFont(font);

    def mousePressEvent(self, event):
        if event.button() == Qt.RightButton:
            # Rewrite the mouse event to a left button event so the cursor is
            # moved to the location of the pointer.
            event = QMouseEvent(QEvent.MouseButtonPress, event.pos(),
                Qt.LeftButton, Qt.LeftButton, Qt.NoModifier)

        QPlainTextEdit.mousePressEvent(self, event)


    def mouseReleaseEvent(self, event):
        if True or event.button() == Qt.LeftButton:
            cursor = self.textCursor()
            c = settings.correction.start(cursor.position())
            if c != -1:
                settings.correction.set_current(settings.correction.from_pos(c))
                self.gui.show_corrections()
            else:
                self.gui.clear_corrections()


    def contextMenuEvent(self, event):
        popup_menu = self.createStandardContextMenu()

        # Select the error
        cursor = self.textCursor()
        c = settings.correction.start(cursor.position())
        has, l = self.highlightCorrection()

        if has:
            text = str(self.textCursor().selectedText())

            action = QtWidgets.QAction("Add \"%s\" to Grammarly Dictionary" % text, self)
            action.setShortcut("Ctrl+L")
            action.triggered.connect(lambda x, w=text: self.add_word(w))
            popup_menu.insertSeparator(popup_menu.actions()[0])
            popup_menu.insertAction(popup_menu.actions()[0], action)


            settings.correction.set_current(settings.correction.from_pos(c))
            self.gui.show_corrections()

            replacements = []
            idx = 0
            while settings.correction.replacement(idx):
                r = settings.correction.replacement(idx)
                if len(text) > 0 and l == 0: r += text[-1]
                replacements.append(r)
                idx += 1


            if len(replacements) > 0:
                popup_menu.insertSeparator(popup_menu.actions()[0])
                i = len(replacements)
                for word in reversed(replacements):
                    action = SpellAction("%s" % word, self)
                    action.correct.connect(self.correctWord)
                    action.setData(word)
                    nfont = action.font()
                    nfont.setBold(True)
                    action.setFont(nfont)
                    action.setShortcut("Ctrl+%d" % (i))
                    popup_menu.insertAction(popup_menu.actions()[0], action)
                    i -= 1
        else:
            self.gui.clear_corrections()

        popup_menu.exec_(event.globalPos())


    def highlightCorrection(self):
        cursor = self.textCursor()
        c = settings.correction.start(cursor.position())
        l = 1
        if c != -1:
            cursor.setPosition(c + 1, QTextCursor.MoveAnchor)
            if settings.correction.end(c) - c < 2: l = 0
            cursor.setPosition(settings.correction.end(c) - l + 1, QTextCursor.KeepAnchor)
            self.setTextCursor(cursor)
            return (True, l)
        return (False, l)


    def correctWord(self, word):
        cursor = self.textCursor()
        cursor.beginEditBlock()

        cursor.removeSelectedText()
        cursor.insertText(word)

        cursor.endEditBlock()


    def add_word(self, word):
        settings.grammarly.add_word(w)




class SpellAction(QAction):
    correct = pyqtSignal(str)

    def __init__(self, *args):
        QAction.__init__(self, *args)

        self.triggered.connect(lambda x: self.correct.emit(
            str(self.data())))


class ParseThread(QThread):
    update_status = pyqtSignal(str)
    update_progress = pyqtSignal(int)
    done = pyqtSignal(int)
    starting = pyqtSignal(int)

    def __init__(self, *args):
        QThread.__init__(self, *args)
        self.gui = args[0]

    def run(self):
        settings.parse(self)
