#!/usr/bin/python
# -*- coding: iso-8859-15 -*-
from enum import Enum
import sys

class LatexPosition(Enum):
    START = 0
    DOCUMENT = 1
    ABSTRACT = 2
    SECTION = 3
    TITLE = 4


class LatexParser:
    
    def __init__(self, latex):
        self.latex = latex
        self._replace = {
            "etal": lambda kw,arg: "et al.",
            "Cref": lambda kw,arg: "Section 1",
            "cref": lambda kw,arg: "section 1",
            "ref": lambda kw,arg: "1",
            "cite": lambda kw,arg: "",
            "begin": lambda kw,arg: "",
            "end": lambda kw,arg: "",
            "label": lambda kw,arg: "",
            "ie": lambda kw,arg: "i.e., ",
            "eg": lambda kw,arg: "e.g., ",
            "SI": self._resolveSI,
            "second": lambda kw,arg: "second",
            "micro\\second": lambda kw,arg: "µs ",
            "milli\\second": lambda kw,arg: "ms ",
            "nano\\second": lambda kw,arg: "ns "
        }    
        self.offsets = [0]
        self.poffsets = [0]
        self.skip = LatexPosition.START
    
    
    def _resolveSI(self, kw, arg):
        ret = ""
        args = arg.split("}{")
        for i in range(len(args)):
            if args[i] == '':
                continue
            if args[i][0] == '\\': ret += self._replacer(args[i][1:], "") 
            else: ret += args[i]
        return ret


    def _replacer(self, kw, arg):
        if kw in self._replace:
            return self._replace[kw](kw, arg)
        else:
            return arg


    def add_replacement(self, kw, fnc):
        self._replace[kw] = fnc
        
        
    def skip_until(self, skip_to):
        self.skip = skip_to
        

    def to_text(self):
        latex = self.latex
        out = ""
        offsets = [i for i in range(len(latex) * 2 + 2)]
        i = 0
        if self.skip != LatexPosition.START:
            while i < len(latex):
                if self.skip == LatexPosition.DOCUMENT and latex[i:].startswith("\\begin{document"):
                    break
                if self.skip == LatexPosition.ABSTRACT and latex[i:].startswith("\\begin{abstract"):
                    break
                if self.skip == LatexPosition.SECTION and latex[i:].startswith("\\section{"):
                    break
                if self.skip == LatexPosition.TITLE and latex[i:].startswith("\\maketitle"):
                    break
                i += 1
                
        while i < len(latex):
            if latex[i] == '~':
                out += " "
                offsets[len(out)] = i
                i += 1
                continue
            if latex[i] == '%':
                while latex[i] != '\n': i += 1
                i += 1
                offsets[len(out)] = i
                continue
            if latex[i] == '\\':
                kw = ""
                i += 1
                while i < len(latex) and latex[i] != ' ' and latex[i] != '{' and latex[i] != '\\' and latex[i] != '.' and latex[i] != ',' and latex[i] != ':' and latex[i] != '!' and latex[i] != ';' and latex[i] != '?' and latex[i] != '-':
                    kw += latex[i]
                    i += 1
                if i >= len(latex): break
                paren = True if latex[i] == '{' else False
                offsets[len(out)] = i
                oldlen = len(out)
                oldstart = i
                if paren:
                    i += 1
                    oldstart += 1
                    arg = ""
                    while len(latex) > i:
                        while len(latex) > i and latex[i] != '}':
                            arg += latex[i]
                            i += 1
                        if len(latex) > (i+1) and latex[i + 1] == '{':
                            arg += latex[i]
                            i += 1
                            continue
                        break
                    if len(latex) > (i+1) and latex[i + 1] == '[':
                        depth = 0
                        while len(latex) > i:
                            i += 1
                            if len(latex) > i and latex[i] == '[': depth += 1
                            if len(latex) > i and latex[i] == ']':
                                i += 1
                                depth -= 1
                                if depth == 0: break
                    #print("%s: %s" % (kw, arg))
                    out += self._replacer(kw, arg)
                else:
                    i -= 1
                    #print("%s" % (kw))
                    out += self._replacer(kw, "")
                
                while oldlen < len(out): 
                    oldlen += 1
                    offsets[oldlen] = oldstart
                    oldstart += 1
                i += 1
                offsets[len(out)] = i
                continue
                    
            if latex[i] == '\n': 
                out += "\n"
                offsets[len(out)] = i
                i += 1
                continue
            
            if len(out) > 0 and out[-1] == ' ':
                if latex[i] == ' ':
                    pass
                elif latex[i] == '.' or latex[i] == ',':
                    out = out[:-1] + latex[i]
                else:
                    out += latex[i]
            else: out += latex[i]
            offsets[len(out)] = i
            i += 1
            
        #print(offsets)
        self.offsets = offsets
        return out
        

    def to_paragraphs(self):
        parsed = self.to_text()
        i = 1
        poffsets = []
        current_start = 0
        paragraphs = []
        while i < len(parsed):
            if parsed[i - 1] == '\n' and parsed[i] == '\n':
                poffsets.append(current_start)
                paragraphs.append(parsed[current_start:i-1])
                current_start = i + 1
            i += 1
        poffsets.append(current_start)
        paragraphs.append(parsed[current_start:])
        self.poffsets = poffsets
        return paragraphs

    
    def original_offset(self, offset, paragraph = 0):
        return self.offsets[self.poffsets[paragraph] + offset]

