from asciimatics.scene import Scene
from asciimatics.screen import Screen
from asciimatics.widgets import Frame, TextBox, Layout, Label, Divider, Text, CheckBox, RadioButtons, Button, PopUpDialog
from asciimatics.exceptions import ResizeScreenError, NextScene, StopApplication, InvalidFields

import sys
import time
import json
import re
import progressbar
import os

import settings

class TUI:
    def __init__(self, screen, corr):
        self.screen = screen
        self.bottom = 10
        self.lines = []
        self.hi_lines = []
        self.text = ""
        self.line_offset = 0
        self.corr = corr
        self.word = ""
        self.write = False
        self.quit_confirm = False


    def set_bottom(self, bottom):
        self.bottom = bottom


    def border(self):
        self.screen.move(0, 0)
        self.screen.draw(self.screen.width - 1, 0, thin=True)
        self.screen.draw(self.screen.width - 1, self.screen.height - self.bottom - 1)
        self.screen.move(self.screen.width, self.screen.height - self.bottom - 1)
        self.screen.draw(-1, self.screen.height - self.bottom - 1, thin=True)
        self.screen.move(0, self.screen.height - self.bottom - 1)
        self.screen.draw(0, 0)
        self.screen.move(0, self.screen.height - 2)
        self.screen.draw(self.screen.width - 1, self.screen.height - 2, thin=True)


    def clear_bottom(self):
        line = " "*self.screen.width
        for b in range(self.bottom):
            self.screen.print_at(line, 0, self.screen.height - self.bottom + b)


    def fast_clear(self):
        line = " " * (self.screen.width + 1)
        for b in range(self.screen.height):
            self.screen.print_at(line, 1, b)


    def parse(self, text):
        self.text = text
        lw = self.screen.width - 4
        pos = 0
        in_highlight = False
        cur_x = 0
        self.lines = []
        self.hi_lines = []
        cur_line = ""
        cur_hi = ""
        highlights = []
        hi_start = 0
        highlight_nr = 0
        delta = 1
        while pos < len(text):
            do_break = False
            for j in range(lw):
                if j >= lw - 1: delta = 1
                if pos >= len(text): break
                if text[pos] == '\n':
                    cur_x = 0
                    self.hi_lines.append(highlights.copy())
                    self.lines.append(cur_line)
                    cur_line = ""
                    highlights = []
                    do_break = True
                    delta = 0
                if self.corr.is_at(pos):
                    end = self.corr.end(pos)
                    if end - 1 == pos: end += 1
                    highlights.append({"start": pos + 1, "end": end - 1, "x": cur_x + delta, "text": text[pos+1:end], "id": highlight_nr, "line": len(self.hi_lines)})
                    highlight_nr += 1
                if text[pos] != '\n': cur_line += text[pos]
                pos += 1
                cur_x += 1
                if do_break:
                    break
            if not do_break:
                self.lines.append(cur_line)
                self.hi_lines.append(highlights.copy())
                highlights = []
                cur_line = ""
                cur_x = 0


    def replace_substr(self, start, length, repl):
        self.text = self.text[0:start+1] + repl + self.text[start+length:]
        self.parse(self.text)


    def wrap(self, text, width):
        out = ""
        cur_len = 0
        for i in range(len(text)):
            if cur_len < width:
                out += text[i]
                if text[i] != '\n':
                    cur_len += 1
                else:
                    cur_len = 0
            else:
                out += text[i] + "\n"
                cur_len = 0
        return out


    def print_html_at(self, text, x, y, color = Screen.COLOUR_WHITE, mode = Screen.A_NORMAL):
        cur_mode = mode
        cur_col = [color]
        pos = 0
        tag = ""
        cur_x = x
        while pos < len(text):
            if text[pos] == "<":
                tag = ""
                while pos < len(text) and text[pos] != ">":
                    pos += 1
                    if pos < len(text): tag += text[pos]
                pos += 1
                tag = tag[0:-1]
                if tag == "b":
                    cur_mode = Screen.A_BOLD
                elif tag == "/b":
                    cur_mode = Screen.A_NORMAL
                elif tag == "/i":
                    cur_col = cur_col[0:-1]
                elif tag == "i":
                    cur_col.append(Screen.COLOUR_CYAN)
            else:
                if pos < len(text):
                    self.screen.print_at(text[pos], cur_x, y, cur_col[len(cur_col) - 1], cur_mode)
                    cur_x += 1
                    pos += 1

    def draw(self):
        for i in range(self.screen.height - self.bottom - 2):
            cur_line = self.lines[i + self.line_offset] if i + self.line_offset < len(self.lines) else ""
            cur_hi = self.hi_lines[i +  self.line_offset] if i + self.line_offset < len(self.hi_lines) else ""
            self.screen.print_at(cur_line, 2, i + 1)
            for c in cur_hi:
                if c["id"] == -1: continue
                self.screen.print_at(c["text"], 2 + c["x"], i + 1, Screen.COLOUR_YELLOW if len(c["text"].strip()) == 0 and c["id"] != self.corr.current_id() else Screen.COLOUR_GREEN, Screen.A_REVERSE if (c["id"] == self.corr.current_id()) or len(c["text"].strip()) == 0 else Screen.A_BOLD)


    def draw_correction(self):
        corrections = self.corr.current_obj()
        if corrections != None:
            xpos = 58 + 4
            self.screen.move(xpos, self.screen.height - self.bottom - 1)
            self.screen.draw(xpos, self.screen.height - 2)
            if "title" in corrections[0]: self.print_html_at(corrections[0]["title"], xpos + 2, self.screen.height - self.bottom, Screen.COLOUR_YELLOW)
            if "explanation" in corrections[0]:
                expl = self.wrap(corrections[0]["explanation"], self.screen.width - (xpos + 4)).split("\n")
                for off,line in enumerate(expl): self.print_html_at(line, xpos + 2, self.screen.height - self.bottom + 1 + off, Screen.COLOUR_WHITE)
            if "replacements" in corrections[0]:
                for idx,r in enumerate(corrections[0]["replacements"]):
                    self.draw_shortcut(idx - 2, str(idx + 1), r)



    def draw_shortcut(self, y, key, desc, xshift = 0):
        self.screen.print_at(" " + key + " ", 1 + xshift, self.screen.height - self.bottom + 2 + y, Screen.COLOUR_WHITE, Screen.A_REVERSE)
        self.screen.print_at(desc, len(key) + 4 + xshift, self.screen.height - self.bottom + 2 + y)


    def draw_menu(self):
        self.draw_shortcut(4, "i", "Ignore")
        self.draw_shortcut(4, "c", "Ignore category", 16)
        self.draw_shortcut(4, "r", "Replace...", 38)
        self.draw_shortcut(5, "x", "Save & Exit")
        self.draw_shortcut(5, "f", "Save & Re-check", 16)

        self.draw_shortcut(5, "a", "Add to dictionary", 38)

        self.draw_shortcut(7, "q/ESC", "Abort (without saving)", self.screen.width - 35)
        self.draw_shortcut(7, "s", "Settings", self.screen.width - 35 - 20)

        if self.write: self.screen.print_at("> " + self.word, 1, self.screen.height - 1, Screen.COLOUR_MAGENTA)


    def draw_settings(self):
        line = " " * (self.screen.width - 20)



    def check_keys(self):
        ev = self.screen.get_key()
        if ev == None:
            return

        if self.quit_confirm:
            if ev in (ord('y'), ord('Y')):
                sys.exit(0)
            elif ev in (ord('n'), ord('N'), ord('q'), ord('Q'), Screen.KEY_ESCAPE):
                self.quit_confirm = False
                self.fast_clear()
                return
            else:
                return

        if self.write:
            if ev == Screen.KEY_ESCAPE:
                self.word = ""
                self.write = False
                self.clear_bottom()
                self.draw_menu()
                self.screen.refresh()
                return
            if int(ev) == ord('\n'):
                self.write = False
                self.corr.replace(self.word, self)
                self.fast_clear()
                self.screen.refresh()
                return
            if int(ev) == Screen.KEY_BACK:
                self.word = self.word[0:-1]
                self.clear_bottom()
                self.screen.refresh()
                return
            if ev >= ord(' ') and ev <= 127:
                self.word += chr(ev)
                return

        move = False
        if ev in (ord('N'), ord('n'), Screen.KEY_RIGHT):
            self.clear_bottom()
            move = True
            self.corr.goto_next()
        elif ev in (ord('B'), ord('b'), Screen.KEY_LEFT):
            self.clear_bottom()
            move = True
            self.corr.goto_prev()
        elif ev in (ord('I'), ord('i')):
            self.corr.remove(self.corr.current_id(), self)
            self.clear_bottom()
            move = True
            self.fast_clear()
        elif ev in (ord('C'), ord('c')):
            self.corr.ignore(self.corr.current_id(), self)
            move = True
            self.fast_clear()

        elif int(ev) >= ord('1') and int(ev) <= ord('9'):
            self.corr.replace(self.corr.replacement(int(ev) - ord('1')), self)
            move = True
            self.fast_clear()

        elif ev in (ord('r'), ord('R')):
            if not self.write:
                self.write = True
                self.word = ""

        elif ev in (ord('Q'), ord('q'), Screen.KEY_ESCAPE):
            self.quit_confirm = True
        elif ev in (Screen.KEY_PAGE_DOWN, Screen.KEY_DOWN):
            self.line_offset += 1
            self.fast_clear()
        elif ev in (Screen.KEY_PAGE_UP, Screen.KEY_UP):
            if self.line_offset > 0: self.line_offset -= 1
            self.fast_clear()

        elif ev in (ord('S'), ord('s')):
            self.screen.play([Scene([SettingsFrame(self.screen)], -1)], stop_on_resize=True)

        elif ev in (ord('X'), ord('x')):
            with open(sys.argv[1], "w") as w:
                w.write(self.text)
            sys.exit(0)

        elif ev in (ord('F'), ord('f')):
            with open(sys.argv[1], "w") as w:
                w.write(self.text)
            self.screen.clear()
            self.screen.refresh()
            main()
            sys.exit(0)

        elif ev in (ord('A'), ord('a')):
            c = self.corr.current_obj()
            if c != None and settings.grammarly != None:
                settings.grammarly.add_word(c[0]["text"])
                self.corr.remove(self.corr.current_id(), self)
                self.clear_bottom()
                move = True
                self.fast_clear()

        if move:
            for line in self.hi_lines:
                for c in line:
                    if c["id"] == self.corr.current_id():
                        if c["line"] > self.line_offset + (self.screen.height - self.bottom - 4):
                            self.fast_clear()
                            self.line_offset = max(0, c["line"] - 4)
                        elif c["line"] < self.line_offset:
                            self.line_offset = max(0, c["line"] - 2)
                            self.fast_clear()
                        break

        if self.quit_confirm:
            self.screen.print_at("Are you sure you want to quit? (y/n)", 1, self.screen.height - 1, Screen.COLOUR_CYAN, Screen.A_BOLD)



class SettingsFrame(Frame):
    def __init__(self, screen):
        super(SettingsFrame, self).__init__(screen,
                                        int(screen.height - 10),
                                        int(screen.width - 10),
                                        data=settings.form_data,
                                        has_shadow=True,
                                        name="Settings")

        layout = Layout([1, 18, 1])
        self.add_layout(layout)
        self._reset_button = Button("Reset", self._reset)
        layout.add_widget(Label("Account:"), 1)
        layout.add_widget(
            Text(label="Username:",
                 name="User",
                 on_change=self._on_change,
                 validator=self._check_email), 1)

        layout.add_widget(
            Text(label="Password:",
                 name="Pwd",
                 on_change=self._on_change), 1)

        layout.add_widget(Divider(height=2), 1)
        layout.add_widget(Label("General Settings:"), 1)
        layout.add_widget(RadioButtons([("American", 1),
                                        ("British", 2)],
                                       label="Dialect:",
                                       name="Dialect",
                                       on_change=self._on_change), 1)

        genre_list = [0, 1, 12, 19, 20, 25, 31, 32]

        layout.add_widget(RadioButtons([(settings.genres[x]["name"], x) for x in genre_list],
                                       label="Genre:",
                                       name="Genre",
                                       on_change=self._on_change), 1)
        for idx,e in enumerate(settings.err_cat):
            layout.add_widget(CheckBox(e["name"],
                                   label="Warn on:" if idx == 0 else "",
                                   name="cat" + str(idx),
                                   on_change=self._on_change), 1)
        layout.add_widget(Divider(height=3), 1)

        layout.add_widget(RadioButtons(settings.parsing_start,
                                       label="Start parsing:",
                                       name="Parsing",
                                       on_change=self._on_change), 1)

        layout.add_widget(Divider(height=3), 1)

        layout2 = Layout([1, 1, 1])
        self.add_layout(layout2)
        layout2.add_widget(self._reset_button, 0)
        layout2.add_widget(Button("Save", self._quit), 2)
        self.fix()


    def _on_change(self):
        changed = False
        self.save()
        for key, value in self.data.items():
            if key not in settings.form_data or settings.form_data[key] != value:
                changed = True
                break
        self._reset_button.disabled = not changed


    def _reset(self):
        self.reset()
        raise NextScene()


    def _quit(self):
        settings.form_data = self.data.copy()
        with open("settings.json", "w") as s:
            json.dump(self.data, s, indent = 2, sort_keys = True)
        raise ResizeScreenError("back")


    @staticmethod
    def _check_email(value):
        m = re.match(r"^[a-zA-Z0-9_\-\.\+]+@[a-zA-Z0-9_\-\.]+\.[a-zA-Z0-9_\-\.]+$",
                     value)
        return len(value) == 0 or m is not None




def app(screen):
    settings.tui = TUI(screen, settings.correction)

    if settings.show_settings:
        screen.play([Scene([SettingsFrame(screen)], -1)], stop_on_resize=True)

    settings.tui.parse(settings.text)
    screen.clear()

    while not screen.has_resized():
        settings.tui.check_keys()
        settings.tui.draw()
        settings.tui.draw_menu()
        settings.tui.draw_correction()
        settings.tui.border()
        settings.tui.draw_settings()
        screen.refresh()
        time.sleep(0.02)
    settings.text = settings.tui.text

 
